#pragma once

#include <string>
#include <unordered_map>
#include <stdexcept>
#include <vector>

namespace ninja
{
    enum class NinjaError
    {
        NONE,
        VARIABLE_ALREADY_DEFINED,
        RULE_ALREADY_DEFINED,
        INVALID_NAME
    };

    class NinjaException : public std::runtime_error
    {
    public:
        NinjaException(const NinjaError _errorCode, const std::string &errMsg = "") : 
            std::runtime_error(errMsg),
            errorCode(_errorCode)
        {

        }

        const NinjaError errorCode;
    };

    struct NinjaVariable
    {
        NinjaVariable(const std::string &name);
        const std::string name;
    };

    class NinjaBuildFile;

    class NinjaArg
    {
    public:
        enum class Type
        {
            NONE,
            VALUE,
            VARIABLE,
            RAW
        };

        NinjaArg() : type(Type::NONE) {}
        NinjaArg(const std::string &value) : type(Type::VALUE), arg(value) {}
        NinjaArg(const NinjaVariable &var) : type(Type::VARIABLE), arg(var.name) {}

        static NinjaArg createRaw(const std::string &value)
        {
            NinjaArg arg;
            arg.type = Type::RAW;
            arg.arg = value;
            return arg;
        }

        Type type;
        std::string arg;
    };

    struct NinjaArgValue
    {
        const NinjaVariable arg;
        const std::string value;
    };

    struct NinjaBuild;

    // Functions throw NinjaException on failure
    class NinjaRule
    {
    public:
        NinjaRule(NinjaBuildFile *buildFile, const std::string &name, const std::string &command);
        NinjaBuild* build(const std::string &in, const std::string &out, const std::vector<ninja::NinjaArgValue> &additionalArgs, const std::vector<NinjaBuild*> &dependsOnBuilds = {});
        NinjaBuild* build(const std::vector<std::string> &in, const std::string &out, const std::vector<ninja::NinjaArgValue> &additionalArgs, const std::vector<NinjaBuild*> &dependsOnBuilds = {});

        const std::string name;
        const std::string command;
        std::string depFile;
        std::string deps;
    private:
        NinjaBuildFile *buildFile;
    };

    struct NinjaBuild
    {
        const NinjaRule *rule;
        const std::vector<std::string> in;
        const std::string out;
        const std::vector<ninja::NinjaArgValue> additionalArgs;
        const std::vector<NinjaBuild*> dependsOnBuilds;
        bool alwaysRun;
    };

    // Functions throw NinjaException on failure
    class NinjaBuildFile
    {
    public:
        ~NinjaBuildFile();

        void defineGlobalVariable(const std::string &name, const std::string &value);
        NinjaRule* createRule(const std::string &name, const std::vector<NinjaArg> &commandArgs);
        NinjaBuild* build(const NinjaRule *rule, const std::string &in, const std::string &out, const std::vector<ninja::NinjaArgValue> &additionalArgs, const std::vector<NinjaBuild*> &dependsOnBuilds = {});
        NinjaBuild* build(const NinjaRule *rule, const std::vector<std::string> &in, const std::string &out, const std::vector<ninja::NinjaArgValue> &additionalArgs, const std::vector<NinjaBuild*> &dependsOnBuilds = {});

        std::string generate() const;
    private:
        std::unordered_map<std::string, std::string> globalVariables;
        std::unordered_map<std::string, NinjaRule*> rules;
        std::vector<NinjaBuild*> builds;
    };
}